/* To output Joe file...
 *
 * Get list of files to pack in
 * With each file, get the file size and add that to the index in memory
 * Add the point that we're at in the output file to the index entry as well, converting it down to the required factor form
 * Copy the file to the output archive
 * After each file, pad with null bytes to the next 4096 block (CurPos % 4096 - 4096 to get number of bytes needed as neg. number?)
 * When all files have been written, write index as well
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "joe.h"

#define MAX_INPUT_LINE 2048 /* hopefully as long as we need */

/* Get data from script, line by line.                                    */
/* Return 0 if business as usual.  -1 if error, 1 if nothing more to read */
int parse_script(FILE* script, char read_filename[], int* read_extra)
{
    char buffer[MAX_INPUT_LINE];
    char number[11];
    int i, j, cap;

    /* Loop on this so we don't have to goto */
    while(1)
    {
        if(!fgets(buffer, MAX_INPUT_LINE, script))
        {
            /* printf("No more lines in script.\n"); */
            return 1;
        }

        i = 0;

        /* Eat leading whitespace */
        while(buffer[i] == ' ' || buffer[i] == '\t')
        {
            ++i;
        }

        /* Loop until we get a line that's relevant                              */
        /* Once it's not a comment or blank line, we can break out.              */
        if (buffer[i] != '#' && buffer[i] != '\n')
            break;
    }

    /* Now get the actually relevant data */

    cap = i + FILENAME_MAX;
    j = 0;

    while(buffer[i] != ' ' && buffer[i] != '\t' && buffer[i] != '\n' && buffer[i] != '\0' && i < cap)
    {
        read_filename[j] = buffer[i];
        ++i;
        ++j;
    }
    read_filename[i] = '\0';

    if (buffer[i] == '\0')
    {
        printf("Malformed script. Each line should be in form <filename> <number>\n");
        return -1;
    }

    while(buffer[i] == ' ' || buffer[i] == '\t')
    {
        ++i;
    }

    cap = i + 10;
    j = 0;

    while(buffer[i] != '\n' && buffer[i] != '\0' && i < cap)
    {
        number[j] = buffer[i];
        ++i;
        ++j;
    }

    number[i] = '\0';

    *read_extra = atoi(number);

    return 0;
}

int append_file(char filename[], FILE* output_file, struct joe_index index[], int idx)
{
   FILE* input_file;
   int i;
   long file_size;
   fpos_t offs;
   int padding;
   int tmp;

   input_file = fopen(filename, "rb");
   if(!input_file)
   {
       printf("Error opening %s\n", filename);
       return -1;
   }

   file_size = 0;

   /* Record the offset we're starting at */
   fgetpos(output_file, &offs);

   index[idx].offset_factor = (long)(offs / 4096);

   while(1)
   {
      fread(&tmp, 1, 1, input_file);
      if(feof(input_file))
         break;
      fwrite(&tmp, 1, 1, output_file);
      ++file_size;
   }

   index[idx].num_bytes = file_size;

   /* Format requires padding to next 4096 multiple */
   /* The result of this formula is negative        */
   fgetpos(output_file, &offs);
   if ((long)offs % 4096)
   {
      padding = file_size % OFFSET_MULTIPLIER - OFFSET_MULTIPLIER;

      for(i = 0; i > padding; --i)
      {
         fputc('\0', output_file);
      }
   }
   return 0;
}

int write_index(FILE* output_file, struct joe_index index[])
{
   int i;
   int tmp;

   for(i = 0; i < MAX_ENTRIES; i++)
   {
      fwrite(&index[i].offset_factor, 4, 1, output_file);
      fwrite(&index[i].num_bytes, 4, 1, output_file);
      fwrite(&index[i].extra, 4, 1, output_file);
   }

   /* Final null values */
   tmp = 0;
   fwrite(&tmp, 4, 1, output_file);

   return 0;
}

int main(int argc, char* argv[])
{
   FILE* list_file;
   FILE* output_file;
   struct joe_index index[MAX_ENTRIES];
   char input_filename[FILENAME_MAX+1];
   int extra_data;
   int file_count = 0;
   int tmp;

   if (argc != 3)
   {
      printf("Usage: JOEPACK <script file> <output file>\n");
      return -1;
   }

   /* List file should be readable as plain text */
   list_file = fopen(argv[1], "r");

   if(!list_file)
   {
      printf("Error: Couldn't open script file.\n");
      return -1;
   }

   output_file = fopen(argv[2], "wb");

   if(!output_file)
   {
      printf("Error: Couldn't prepare output file for writing.\n");
      return -1;
   }


   /* Make sure there is no garbage in the index */
   int i;
   for(i = 0; i < MAX_ENTRIES; i++)
   {
      index[i].offset_factor = 0;
      index[i].num_bytes = 0;
      index[i].extra = 0;
   }


   while(1)
   {
       tmp = parse_script(list_file, input_filename, &extra_data);
       if(tmp != 0)
           break;
       /* FIXME: This is just a debug print to make sure everything's good so far */
       /* printf("Read line: %s %d\n", input_filename, extra_data);               */
       index[file_count].extra = extra_data;
       tmp = append_file(input_filename, output_file, index, file_count);
       if(tmp != 0)
       {
          printf("Error adding %s - I will not continue.\n", input_filename);
          return -1;
       }
       file_count++;
   }

   write_index(output_file, index);

   fclose(output_file);

   return 0;
}
